
/*
	File: Array.h
	(c) Oscar Nierstrasz 30.3.96

	A generic Array class. Supports dynamic resizing.
	Can be used to implement a stack.
*/

#ifndef ARRAY_H
#define ARRAY_H

#include <exception.h>
#include <iostream.h>

// #define DEBUG
#include "Debug.h"

template <class Item>
class Array
{
public:
	Array (int initSize=0);

	Array (int initSize, const Item& initVal);	
	Array (const Array<Item>& copy);
	
	Array (int size, const Item* items);

	~Array (void);

	Array<Item>& operator= (const Array<Item>&) throw(xmsg);

	Item& operator[] (const int n) throw(xmsg) const;
	int size (void) throw(xmsg) const { return curSize; }
	
	// should we provide an optional initVal if newSize is larger?
	void resize (int newSize) throw(xmsg);

protected:
	int curSize;
	int maxSize;
	Item * arrayRep;

	// Helper functions:
	Item * makeArray(int);
	void become(int initSize, int n, const Item* values);
};

// default constructor
template <class Item>
Array<Item>::Array(int initSize) :
	curSize(initSize),
	maxSize(initSize),
	arrayRep(makeArray(maxSize))
{
	debug << "Array::Array(int) -- Made new array. curSize = " << curSize;
	debug << ", maxSize = " << maxSize << endl;	
}

// initializing constructor
template <class Item>
Array<Item>::Array(int initSize, const Item& initVal) :
	curSize(initSize),
	maxSize(initSize),
	arrayRep(makeArray(maxSize))
{
	debug << "Array::Array(int) -- Made new array. curSize = " << curSize;
	debug << ", maxSize = " << maxSize << endl;
	for (i=0; i<curSize; i++)  {
		arrayRep[i] = initVal;
	}
}

// copy constructor
template <class Item>
Array<Item>::Array(const Array<Item>& copy) :
	arrayRep(0)
{
	debug << "Array::Array(Array) -- copy constructor called" << endl;
	become(copy.maxSize, copy.curSize, copy.arrayRep);
}

// array constructor
template <class Item>
Array<Item>::Array(int size, const Item* items) :
	arrayRep(0)
{
	debug << "Array::Array(int,Item*) -- constructor called" << endl;
	become(size, size, items);
}

template <class Item>
Array<Item>::~Array(void)
{
	debug << "Array::~Array -- destructor called" << endl;
	delete [] arrayRep;
}

template <class Item>
Item&
Array<Item>::operator[](const int n) const
{
	if ((n<0) || (n >= curSize)) {
		throw(xmsg("Array::operator[]: index out of range"));
	}
	return arrayRep[n];
}

// assignment
template <class Item>
Array<Item>&
Array<Item>::operator=(const Array<Item>& val)
{
	debug << "Array::operator= called" << endl;
	
	if (this != &val) {
		become(val.maxSize, val.curSize, val.arrayRep);
	}
	return *this; // MPW caught this
}

template <class Item>
void Array<Item>::resize (int newSize)
{
	if (newSize < 0) {
		throw(xmsg("Array::resize: cannot take negative size"));
	}
	debug << "Array::resize from " << curSize << " to " << newSize << endl;
	if (newSize <= maxSize) {
		// if we are too small now, should reclaim space
		if (maxSize > 4*newSize) {
			become(newSize, newSize, arrayRep);
		} else {
			curSize = newSize;
		}
		return;
	} else {
		// newSize > maxSize, so become need to become larger
		int newMax = (maxSize>0) ? maxSize : 2; // start positive!
		while (newSize > newMax) {
			newMax = 2*newMax; // expect to double in size ...
		}
		become(newMax, curSize, arrayRep); // copy current state
		curSize = newSize; // and now we have room to resize
		return;
	}
}

// HELPER FUNCTIONS

// requires: arrayRep is initialized to some array (or to 0)
// ensures: consistent representation
template <class Item>
void
Array<Item>::become(int availSize, int newSize, const Item* values)
{
	if (newSize > 0 && values == 0) {
		// private helper function, so this should never happen!
		throw(xmsg("Array::become: newSize>0 but values=0!"));
	}

	// assume availSize >= 0
	if (availSize < newSize) {
		availSize = newSize; // sanity check
	}

	Item * oldRep = arrayRep;	// save just in case values overlaps arrayRep!
	arrayRep = makeArray(availSize); // NB: this might fail
	maxSize = availSize;
	curSize = newSize;
	for (int i=0; i<newSize; i++) {
		arrayRep[i] = values[i];		
	}
	delete [] oldRep;
	debug << "Array::become -- Became a new array. maxSize = " << maxSize;
	debug << ", curSize = " << curSize << endl;
}

template <class Item>
Item *
Array<Item>::makeArray(int size)
{
	debug << "Array::makeArray -- Making a new array, avail size = " << size << endl;
	if (size < 0) {
		throw(xmsg("Array::makeArray: Can't allocate negative sized array!"));
	}
	Item * newArray = new Item[size]; // NB: Call Item constructor size times
	if (newArray == 0) {
		throw(xmsg("Array::makeArray: Couldn't allocate enough space"));
	}
	return newArray;
}

#endif // ARRAY_H
